#include "uiribbon.inc"
#include "util.inc"
#include "xml.inc"
#autodefine "OFF"


// base class for all COM classes
class CUnknown
{
	declare CUnknown();
	declare virtual QueryInterface(REFIID *riid, void *ppvObject),HRESULT;
	declare virtual AddRef(),ULONG;
	declare virtual Release(),ULONG;
	int m_refcnt;
	QITAB *m_quitab; // not initialized in Unknown::Unknown()
}

// class exported from DllGetClassObject
class ClassFactory : CUnknown
{
	declare ClassFactory();
	declare virtual CreateInstance(IUnknown *pUnkOuter, REFIID riid, void *ppvObject),HRESULT;
	declare virtual LockServer(BOOL fLock),HRESULT;

	CLSID *m_pClsid; // set in DllGetClassObject
}



// provides the core functionality for the Ribbon or ContextualUI Views of the Framework
class CUIFramework : CUnknown
{
	// _CLSID_UIRibbonFramework
	// _IID_IUIFramework
	declare CUIFramework();
	declare virtual Release(),ULONG;
	declare virtual Initialize(HWND frameWnd,IUIApplication* application),HRESULT;
	declare virtual Destroy(),HRESULT;
	declare virtual LoadUI(HINSTANCE instance,LPCWSTR resourceName),HRESULT;
	declare virtual GetView(DWORD viewId,IID *riid,void* ppv),HRESULT;
	declare virtual GetUICommandProperty(DWORD commandId,PROPERTYKEY* key,/*out*/PROPVARIANT* value),HRESULT;
	declare virtual SetUICommandProperty(DWORD commandId,PROPERTYKEY* key,VARIANT* value),HRESULT;
	declare virtual InvalidateUICommand(DWORD commandId,UI_INVALIDATIONS flags,PROPERTYKEY* key),HRESULT;
	declare virtual FlushPendingInvalidations(),HRESULT;
	declare virtual SetModes(int iModes),HRESULT;

	// sub DuplicateXmlNode(XML *node, XML *ppv),BOOL;
	declare RegisterRibbonClass();
	declare SimulationThread();
//	declare EnumCommandsInView(XML *views, UI_COMMANDTYPE type, LPWSTR pwszName);
	declare FindCommandByCommandName(XML *node, COMMANDDESC *ppv),BOOL; // find node->CommandName in m_commands
	declare GetCommandLabelTitleString(XML *command,/*out*/XML *LabelTitleString),BOOL;
	declare CleanupHandler(UICOMMAND *puicmd);
	// GDI
	declare InitGDI();
	declare CleanupGDI();
	declare OnPaint();
	declare Redraw();
	declare QueueMouseLeave();
	// commands
	declare LoadCommands(XML *application);
	declare LoadCommand(XML *command, int nCmdID byref, HDC hdc);
	declare LookupCommandByCommandId(int nCmdID, COMMANDDESC *ppv),BOOL;
	declare ExecuteCommand(UICOMMAND *command, PROPVARIANT *property);
	// views
	declare LoadViews(XML *application);
	declare LoadView(XML *view, int nViewID byref);
	// tabs
	declare CleanupTabDesc(TABDESC *tdesc);
	declare GetCurrentTab(TABDESC *ppv),BOOL;
	declare GetTabByIndex(int index, TABDESC *ppv),BOOL;
	declare MeasureTabLabel(TABDESC *tabdesc, LPWSTR pwszText);
	declare GetTabButtonRect(int index, RECT *rc),BOOL;
	declare GetTabIndexFromPoint(POINT *pt, int *ppv),BOOL;
	declare DrawTabs(HDC hdc, RECT *prc);
	declare HighlightTab(int index, BOOL fVisualize),BOOL;
	// groups
	declare LoadGroup(TABDESC *tabdesc, XML *group, int left byref);
	declare CleanupGroupDesc(GROUPDESC *gdesc);
	declare MeasureGroup(GROUPDESC *gdesc);
	declare DrawGroups(HDC hdc, RECT *prc);
	declare DrawGroup(HDC hdc, GROUPDESC *gdesc, int nGroupIndex, RECT *rcGroup);
	declare GetGroupRect(TABDESC *tdesc, int index, RECT *prc),BOOL;
	declare GetGroupIndexFromPoint(POINT *pt, int *ppv),BOOL;
	declare GetGroupByIndex(TABDESC *tdesc, int index, GROUPDESC *ppv),BOOL;
	declare HighlightGroup(int index, BOOL fVisualize),BOOL;
	// controls
	declare GetControlFromPoint(POINT *pt, CONTROLDESC *ppControl),BOOL;
	declare HighlightControl(CONTROLDESC *control, BOOL fVisualize),BOOL;

	// events
	declare OnTabSelChanged();

	// frame window
	HWND       m_frameWnd;
	WNDPROC    m_wndproc;
	// ribbon window
	HWND       m_hwnd;
	SIZE       m_size;      // width/height
	XML        m_markup;
	IUIRibbon *m_RibbonView;
//	POINT      m_ptCursor;
	// Initialize
	IUIApplication* m_application;
	// LoadUI
	HINSTANCE m_instance;
	//LPCWSTR m_resourceName;
	CPointerlist m_commands; // COMMANDDESC

	//-------------------------------------------- TABS
	CPointerlist m_tabs;          // TABDESC
	IUICommandHandler *m_tabshandler;
	int          m_ActiveTab;     // selected tab index, or -1
	int          m_TabHeight;     // tab buttons height
	int          m_HotTab;        // highlighted tab index, or -1
	HFONT        m_TabFont;       // stock font
	HPEN         m_LinePen1;      // deleted in CleanupGDI()
	HPEN         m_LinePen2;      // deleted in CleanupGDI()
	HPEN         m_LinePen3;      // deleted in CleanupGDI()
	HPEN         m_LinePen4;      // 
	HBRUSH       m_hbrBkgnd;      // deleted in Release()
	HBRUSH       m_hbrLabel;      // deleted in CleanupGDI()
	BITMAPDATA   m_bmdTabBk;
	BITMAPDATA   m_bmdTabBkHot;
	BITMAPDATA   m_bmdTabMenuIcon;

	//-------------------------------------------- TABS

	COLORREF     m_TabFontColor;

	//-------------------------------------------- RIBBON

	BITMAPDATA   m_bmdRibbonGradient;

	//-------------------------------------------- GROUPS

	BITMAPDATA   m_bmdGroupDivider; // must be loaded before LoadView()
	BITMAPDATA   m_bmdGroupHot;
	HFONT        m_GroupLabelFont;// stock font
	int          m_HotGroup;      // highlighted group index, or -1
	COLORREF     m_GroupLabelColor;

	//-------------------------------------------- BUTTONS

	BITMAPDATA   m_bmdButtonHotGradient;
	BITMAPDATA   m_bmdButtonPushGradient;
	HFONT        m_ButtonFont;// stock font
	COLORREF     m_ButtonFontColor;
	HPEN         m_ButtonBorderPen;
	pointer      m_pHotControl;        // CONTROLDESC*

	//-------------------------------------------- SIMULATION
	HANDLE m_hEvent;
	HANDLE m_hThread;
}

// Application -> Application.Commands -> Command list
struct COMMANDDESC // CleanupCommandDesc()
{
	DWORD       commandId;
	XML        *xmlnode;
	// images
	BITMAPDATA  bmdSmallImage;
	BITMAPDATA  bmdLargeImage;
	// text
	BSTR        bstrLabelTitle;
	int         nTextWidth;    // CalculateTextWidths(HDC hdc, LPWSTR pwsztext, int *ppTextWidth, int *ppMaxWordWidth);
	int         nMaxWordWidth;
}

const BTNWIDTHLARGE = 32;
const BTNWIDTHSMALL = 16;

struct UICOMMAND // CleanupHandler
{
	IUICommandHandler *handler;
	UI_COMMANDTYPE     type;
	COMMANDDESC       *command;
}

struct TABDESC // CleanupTabDesc()
{
	UICOMMAND     uicmd;
	XML          *TabNode;  // <Tab>
	SIZE          LabelSize;
	CPointerlist *groups;   // GROUPDESC
	BOOL          fIsMenu;
	LPWSTR        pwszText; // static text [temporary]
}

enum LAYOUTTEMPLATE
{
	OneButton,
	TwoButtons,
	ThreeButtons,
	ThreeButtonsOneBigAndTwoSmall,
	ThreeButtonsAndOneCheckBox,
	FourButtons,
	FiveButtons,
	FiveOrSixButtons,
	SixButtons,
	SixButtonsTwoColumns,
	SevenButtons,
	EightButtons,
	EightButtonsLastThreeSmall,
	NineButtons,
	TenButtons,
	ElevenButtons,
	OneFontControl,
	OneInRibbonGallery,
	InRibbonGalleryAndBigButton,
	InRibbonGalleryAndButtonsGalleryScalesFirst,
	ButtonGroups,
	ButtonGroupsAndInputs,
	BigButtonsAndSmallButtonsOrInputs
}

enum CTLTYPE
{
	TYPE_Button,
	TYPE_CheckBox,
	TYPE_ComboBox,
	TYPE_ControlGroup,
	TYPE_DialogLauncher,
	TYPE_DropDownButton,
	TYPE_DropDownColorPicker,
	TYPE_DropDownGallery,
	TYPE_FontControl,
	TYPE_InRibbonGallery,
	TYPE_Spinner,
	TYPE_SplitButton,
	TYPE_SplitButtonGallery,
	TYPE_ToggleButton
}

struct GROUPDESC // CleanupGroupDesc()
{
	UICOMMAND     uicmd;
	TABDESC     *OwningTab;
	XML         *GroupNode;
	int          left;
	int          width;
	LAYOUTTEMPLATE SizeDefinition;
	CPointerlist  *controls; // CONTROLDESC
}

enum SIZETEMPLATE
{
	SIZETEMPLATE_LARGEBUTTON,
	SIZETEMPLATE_SMALLBUTTON
}
class CONTROLDESC
{
	UICOMMAND     m_uicmd;
	CUIFramework *m_framework;
	CTLTYPE       m_type;
	GROUPDESC    *m_group;
	XML          *m_xmlNode;
	RECT          m_rc;
	declare _CONTROLDESC();
	declare SetSizeTemplate(SIZETEMPLATE type);
	declare Width(),int;
	declare SetPos(int x, int y); // client coordinates
	declare SetSize(int w, int h);
	declare Draw(HDC hdc);
	//declare OnMouseMove(POINT pt);
	declare OnLButtonDown();
	declare OnLButtonUp(BOOL clicked);
	DWORD   m_state;
}

// defines the method for retrieving an image for display in the Ribbon user interface (UI)
class CUIImage : CUnknown
{
	// _IID_IUIImage
	declare CUIImage();
	declare virtual Release(),ULONG;
	declare virtual GetBitmap(/*out*/HBITMAP* bitmap),HRESULT;

	HBITMAP      m_bitmap;
	UI_OWNERSHIP m_options;
}

// defines the method for creating an IUIImage object
class CUIImageFromBitmap : CUnknown
{
	// _CLSID_UIRibbonImageFromBitmapFactory
	// _IID_IUIImageFromBitmap
	declare CUIImageFromBitmap();
	declare virtual CreateImage(HBITMAP bitmap,UI_OWNERSHIP options,/*out*/CUIImage* ppimage),HRESULT;
}

// provides the ability to specify settings and properties for a Ribbon
class CUIRibbon : CUnknown
{
	// _IID_IUIRibbon
	declare CUIRibbon();
	declare virtual Release(),ULONG;
	declare virtual GetHeight(/*out*/DWORD* cy),HRESULT;
	declare virtual LoadSettingsFromStream(IStream* pStream),HRESULT;
	declare virtual SaveSettingsToStream(IStream* pStream),HRESULT;

	int           m_nViewID;
	CUIFramework *m_framework;
}

// provides the core functionality for a ContextPopup.
class CUIContextualUI : CUnknown
{
	// _IID_IUIContextualUI
	declare CUIContextualUI();
	declare virtual Release(),ULONG;
	declare virtual ShowAtLocation(int x,int y),HRESULT;

	int           m_nViewID;
	CUIFramework *m_framework;
}

// defines the methods for manipulating collection-based Gallery controls at run time,
// including the Application Menu and Quick Access Toolbar
// TODO
class CUICollection : CUnknown
{
	// _IID_IUICollection
	declare CUICollection();
	declare virtual Release(),ULONG;
	declare virtual GetCount(/*out*/DWORD* count),HRESULT;
	declare virtual GetItem(DWORD index,/*out*/IUnknown* ppitem),HRESULT;
	declare virtual Add(IUnknown* item),HRESULT;
	declare virtual Insert(DWORD index,IUnknown* item),HRESULT;
	declare virtual RemoveAt(DWORD index),HRESULT;
	declare virtual Replace(DWORD indexReplaced,IUnknown* itemReplaceWith),HRESULT;
	declare virtual Clear(),HRESULT;
}
// TODO
class CConnectionPoint : CUnknown
{
	declare CConnectionPoint();
	declare virtual GetConnectionInterface(/*out*/IID *pIID),HRESULT;
	declare virtual GetConnectionPointContainer(/*out*/IConnectionPointContainer *ppCPC),HRESULT;
	declare virtual Advise(IUnknown *pUnkSink,/*out*/DWORD *pdwCookie),HRESULT;
	declare virtual Unadvise(DWORD dwCookie),HRESULT;
	declare virtual EnumConnections(/*out*/IEnumConnections *ppEnum),HRESULT;
}
// TODO
class CEnumConnectionPoints : CUnknown
{
	declare CEnumConnectionPoints();
	declare virtual Next(ULONG cConnections,/*out*/CConnectionPoint *ppCP,/*out*/ULONG *pcFetched),HRESULT;
	declare virtual Skip(ULONG cConnections),HRESULT;
	declare virtual Reset(),HRESULT;
	declare virtual Clone(/*out*/CEnumConnectionPoints *ppEnum),HRESULT;
}
// TODO
class CConnectionPointContainer : CUnknown
{
	declare CConnectionPointContainer();
	declare virtual EnumConnectionPoints(/*out*/CEnumConnectionPoints *ppEnum),HRESULT;
	declare virtual FindConnectionPoint(REFIID* riid,/*out*/CConnectionPoint *ppCP),HRESULT;
}


